unit ModDisplay;
interface

uses ModMain, Windows, SysUtils, Classes,
     UVideoRendererClass, UVideoRConst,
     UvideoRendererOverlay,UvideoRendererRGB;

const VideoSize=4096*256;

type
  TBGR        = Record b,g,r:Byte End;
  TPalette    = Array[0..255] of TBGR;

  TYUV        = Record Y,U,V:Byte end;
  TYUVPalette = Array[0..255] of TYUV;

  TPalBitmapInfo=record
   bmiHeader:TBitmapInfoHeader;
   bmiColors:Array[0..15] of Word;
  end;

  TLine8 = array [Word] of byte;
  PLine8 = ^TLine8;
  TLine16 = array [Word] of Word;
  PLine16 = ^TLine16;
  Tline24 = Array [Word] of TBGR;
  PLine24 = ^TLine24;
  TLine32 = array [Word] of LongWord;
  PLine32 = ^TLine32;

  TMode       = (GDI,DDRAW,OVL);

  TOutput=class
   private
    fHwnd: HWND;

    fRenderer:TVideoRendererClass;

    fOVLCol:TYuvPalette;      // Overlay-YUY2-Palette
    fPalette:TPalette;

    bmpPAL:HPalette;
    bmpInfo:TPalBitmapInfo;

    fSpeccyline: array[0..191] of word;
    fFlashstat:Boolean;
    ftop,fleft: Word;
    fBorder: Word;

    fReverse_attr:array[0..127] of byte;
    fVga_h_nib:array[0..15,0..127] of longint;
    fMode: TMode;
    fSkip,Skipped: Byte;
    fFlash:Boolean;
    fDesktopMode:Boolean;

    procedure SetPalette(Pal:TPalette);

    procedure AdjustDim(BorderEmu:Boolean);
    procedure MakeTable;

    procedure calcovlpal;
    procedure PutPixel(X,Y:DWord;Color:Word);
    procedure putYUVpixel(x,y:DWord;Color:word);
    procedure putRGBpixel(x,y:DWord;Color:word);

    procedure SetDesktopMode(Value:Boolean);
   public
    videomemory: pointer; // Videodata

    constructor create(Handle:HWND;Borderemu:Boolean;Mode:TMode);
    destructor Destroy;

    procedure PaintColorKey(Rect:TRect);

    procedure Setborder(Y:Integer);
    procedure Updateline(Y:Integer);
    procedure Repaint(R:TRect);

    property Palette: TPalette read fPalette write SetPalette;

    property Left: Word read FLeft; // Borderwidth
    property Top: Word read FTop;   // Borderheight
    property Flash_Status: Boolean read fFlashStat write fFlashStat;
    property Border: Word read fBorder write FBorder;
    property Skip: Byte read fSkip write fSkip;

    property Mode: TMode read FMode;
    property Flash: Boolean read fFlash write FFlash;

    property DesktopMode: Boolean read fDesktopMode write SetDesktopMode;
  end;

  const YaseCol:array[0..15] of TBGR =(
        (B:000;G:000;R:000),
        (B:160;G:000;R:000),
        (B:000;G:000;R:220),
        (B:180;G:000;R:228),
        (B:000;G:212;R:000),
        (B:212;G:212;R:000),
        (B:000;G:208;R:208),
        (B:200;G:200;R:200),
        (B:000;G:000;R:000),
        (B:172;G:000;R:000),
        (B:000;G:000;R:240),
        (B:220;G:000;R:252),
        (B:000;G:240;R:000),
        (B:252;G:252;R:000),
        (B:000;G:252;R:252),
        (B:252;G:252;R:252));

var
  display: TOutput;

implementation

procedure toutput.SetDesktopMode(Value:Boolean);
begin
 if fDesktopmode=Value then exit;
 fDesktopmode:=Value;
 frenderer.SetDesktopPlay(fDesktopMode);
end;


procedure toutput.maketable;
var v,x,y,z:integer;
    fore_attrib,back_attrib:array[0..255] of byte;
    qtemp:Quadruple;
    Pal:TPalette;
begin
 // Create Table
 v:=0;
 for x:=0 to 2 do
   for y:=0 to 7 do
     for z:=0 to 7 do
      begin
       fSpeccyline[v]:=x shl 11+y shl 5+z shl 8;
       inc(v);
      end;

 for x:=0 to 255 do
  begin
   y:=x and 7; (* fore *)
   z:=(x and 56) shr 3; (* back *)
   if y=0 then y:=y or 8;
   if z=0 then z:=z or 8;

   if x and 64>0 then
    begin
     y:=y or 8;
     z:=z or 8;
    end;
    fore_attrib[x]:=y;
    back_attrib[x]:=z;
  end;

 for x:=0 to 15 do
  begin
   for y:=0 to 127 do
    begin
     if x and 1>0 then
      qtemp.b4:=fore_attrib[y]
     else
      qtemp.b4:=back_attrib[y];

     if x and 2>0 then
      qtemp.b3:=fore_attrib[y]
     else
      qtemp.b3:=back_attrib[y];

     if x and 4>0 then
      qtemp.b2:=fore_attrib[y]
     else
      qtemp.b2:=back_attrib[y];

     if x and 8>0 then
      qtemp.b1:=fore_attrib[y]
     else
      qtemp.b1:=back_attrib[y];
      fvga_h_nib[x][y]:=qtemp.Q;
    end;
  end;

 for x:=0 to 127 do
  begin
   y:=(x and 7) shr 3;
   z:=(x and 56) shr 3;
   fReverse_attr[x]:=(x and 64) or y or z;
  end;

 For X:=low(YaseCol) to high(YaseCol) do Pal[X]:=YaseCol[X];
 Palette:=Pal;
end;

procedure toutput.PaintColorKey(Rect:TRect);
begin
 fillchar(videomemory^,videosize,0);

 if not assigned(frenderer) then exit;
 frenderer.ClearToBlack;
 frenderer.UpdateFrame(rect);
end;

procedure toutput.AdjustDim(Borderemu:Boolean);
begin
  if borderemu then
  begin ftop:=32;fleft:=32; end else
  begin ftop:=0;fleft:=0; end;
end;

procedure toutput.calcovlpal;

 function f2b(f:double):byte;
 var i:integer;
 begin
  i:=round(f);
  if i>255 then i:=255 else if i<0 then i:=0;
  result:=i;
 end;

var x:integer;
begin
 for x:=low(Palette) to High(Palette) do
   with Palette[x],fovlcol[x] do
     begin
        Y:= f2b(0.299*R + 0.587*G + 0.114*B);
        U:= f2b(128.0 - 0.1684*R - 0.3316*G + 0.5*B);
        V:= f2b(128.0 + 0.5*R - 0.4187*G - 0.0813*B);
      end;
end;

constructor toutput.create(Handle:HWND;Borderemu:Boolean;Mode:TMode);
var i:integer;
begin
 Adjustdim(Borderemu);
 case mode of
 GDI: begin
       MakeTable;
       with BmpInfo do begin
       with bmiheader do begin
        bisize:=sizeof(bmiheader){40};
        biheight:=192+ftop shl 1;
        biwidth:=256+fleft shl 1;
        //videosize:=biWidth*biHeight;
        biplanes:=1;
        bibitcount:=8;
        bicompression:=BI_RGB;
        bisizeimage:=0;
        bixpelspermeter:=0;
        biypelspermeter:=0;
        biclrused:=16;
        biclrimportant:=0;
       end;
       for i:=0 to 15 do bmiColors[i]:=i;
      end;
      getmem(videomemory,videosize);
     end;
 ovl,
 DDraw: begin
         if mode=DDraw then
          frenderer:=TVideoRendererRGB.Create else
          frenderer:=TVideoRendererOverlay.create;
         frenderer.InitDDraw(handle,256+fleft shl 1,192+ftop shl 1);
         //videosize:=frenderer.width*frenderer.height*frenderer.bpp;
         getmem(videomemory,videosize);
         Maketable;
         CalcOvlPal;
       end;
 end;
 fHwnd:= Handle;
 fMode:=mode;
end;

destructor tOutPut.Destroy;
begin
 if bmppal<>0 then DeleteObject(bmpPAL);
 SetDesktopMode(false);
 if assigned(frenderer) then frenderer.destroy;
 if assigned(videomemory) then freemem(videomemory,4096*256);
 videomemory:=nil;
end;

procedure TOutPut.SetPalette(Pal:TPalette);
var
  col: PLogPalette;
  i: Integer;
begin
 if not assigned(self) then exit;
 fpalette:=pal;calcOvlPal;

  col := nil;
  try
    GetMem(col, sizeof(TLogPalette) + sizeof(TPaletteEntry) * 255);
    col.palVersion := $300;
    col.palNumEntries := 256;
    for i := low(Pal) to high(Pal) do
    with Pal[i] do
    begin
      col.palPalEntry[i].peRed := r;
      col.palPalEntry[i].peGreen := g;
      col.palPalEntry[i].peBlue := b;
    end;
    bmppal := CreatePalette(col^);
  finally
    FreeMem(col);
  end;
end;

function getbit(Color:Longint;B:Byte):longint;
var q:Quadruple absolute Color;
begin
  case b of
  0: result:=q.b1;
  1: result:=q.b2;
  2: result:=q.b3;
  3: result:=q.b4;
 end;
end;

procedure toutput.putRGBpixel(x,y:DWord;Color:word);
var offset:DWord;
    C: LongWord;
begin
 offset:=(x+y*frenderer.width);
 with fPalette[color] do
  c:=getrgb(r,g,b,frenderer.GetVideoMode);

 case frenderer.GetVideoMode of
  VIDEOMODE_RGB16:pline16(videomemory)^[offset]:=c;
  VIDEOMODE_RGB24:pline24(videomemory)^[offset]:=fPalette[color];
  VIDEOMODE_RGB32:pline32(videomemory)^[offset]:=c;
 end;

end;

procedure toutput.PutPixel(X,Y:DWord;Color:Word);
begin
 if mode=ovl then putYuvPixel(X,Y,Color) else PutRGBPixel(X,Y,Color);
end;

procedure toutput.putYUVpixel(x,y:DWord;Color:word);
var offset:DWord;
    C: pair;
begin
 offset:=(x+y*frenderer.width);
 with fovlcol[color] do
 begin
  if frenderer.GetVideoMode=VIDEOMODE_YUY2 then
  begin
   if not odd(x) then c.h:=U else c.H:=V; c.L:=Y;
  end else
  begin
   if not odd(x) then c.l:=U else c.l:=V; c.H:=Y;
  end;
 end;
 pline16(videomemory)^[offset]:=c.w;
end;

procedure TOutPut.Setborder(Y:Integer);
var color,xmax:longint;
    x:integer;
begin
 xmax:=256+fleft shl 1;
 case fmode of
 GDI: with BmpInfo.bmiHeader do
       fillchar(pline8(videomemory)^[(biheight-y)*biWidth],xmax,border);
  OVL,DDraw: for x:=0 to xmax do putpixel(x,y,border);
 end;
end;

procedure TOutput.updateline(Y:Integer);
var bitmap, farbepos:word;
    i, x,color, attr, data:Byte;
    offs:Word;
    value:Longint;
    scanline:Pline8;
begin
 if not assigned(self) then exit;
   if fmode=gdi then
    with BmpInfo.bmiHeader do
     Scanline:=pline8(dword(videomemory)+(biheight-(y+ftop))*biwidth);

   if (fleft>0) or (ftop>0) then // Realborder
   begin
    if (y<0) or (y>191) then begin setborder(y+ftop);exit;end;
    case fmode of
     gdi: begin
            fillchar(scanline^[0],fleft,border);
            fillchar(scanline^[256+fleft],fleft,border);
          end;
     OVL,DDraw: for offs:=0 to fleft-1 do
             begin
              putpixel(offs,y+ftop,border);
              putpixel(offs+256+fleft,y+ftop,border);
             end;
    end;//case
   end;
   bitmap:=fSpeccyline[y];
   offs:=fleft;
   farbepos:=6144 or ((y and 248) shl 2);

   for x:=0 to 31 do
    begin
     color:=gRAMPage[glUseScreen, farbepos];inc(farbepos);
     attr:=color and 127;

     if flash then
     begin
      if (color and 128>0) and (fFlashstat) then
       attr:=fReverse_attr[attr];
     end;

      data:=gRAMPage[glUseScreen, bitmap];inc(bitmap);

      case fmode of
       GDI: begin
             move(fVga_h_nib[(data and $F0) shr 4][attr],scanline^[offs],4);inc(offs,4);
             move(fVga_h_nib[data and $0F][attr],scanline^[offs],4);inc(offs,4);
            end;
       OVL,DDRAW: begin
               value:=fVga_h_nib[(data and $F0) shr 4][attr];
               for i:=0 to 3 do putpixel(offs+i,y+ftop,getbit(value,i));
               inc(offs,4);
               value:=fVga_h_nib[data and $0F][attr];
               for i:=0 to 3 do putpixel(offs+i,y+ftop,getbit(value,i));
               inc(offs,4);
              end;
      end;
    end;
end;

procedure TOutput.Repaint; // Lame :)
var dc:Hdc;
begin
if not assigned(self) then exit;
 if fskip>0 then
  begin
   skipped:=(skipped+1) mod fskip;
   if skipped>0 then exit;
  end;
 case fMode of
  GDI: with BmpInfo.bmiHeader do begin
       dc:=getdc(fhwnd);
       SelectPalette(DC,bmpPAL,False);
       RealizePalette(DC);
       StretchDIBits(
         DC,	// handle of device context
          r.left,	// x-coordinate of upper-left corner of dest. rect.
          r.top,	// y-coordinate of upper-left corner of dest. rect.
          r.Right,	// width of destination rectangle
          r.bottom,	// height of destination rectangle
          0,	// x-coordinate of upper-left corner of source rect.
          0,	// y-coordinate of upper-left corner of source rect.
          biWidth,	// width of source rectangle
          biHeight,	// height of source rectangle
          videomemory,	// address of bitmap bits
          pBitmapInfo(@bmpInfo)^,	// address of bitmap data
          dib_pal_colors,	// usage
          SRCCOPY 	// raster operation code
          );
        releaseDC(fhwnd,dc);
       end;
  DDRAW,OVL: if assigned(frenderer) then frenderer.Draw(videomemory,rect(0,0,0,0),0,1);
 end;
end;

end.
