unit UVideoRendererOverlay;

interface

uses Windows, SysUtils, DirectDraw, DirectShow, ActiveX,
     UVideoRConst, ShlObj, UVideoRendererClass;

const
  OVERLAY_COLORKEY_16 = $00000001;
  OVERLAY_COLORKEY_32 = $0000000F;
  OVERLAY_COLOR       = $000F0000;
  IID_IActiveDesktop : TGUid = '{F490EB00-1240-11D1-9888-006097DEACF9}';

type
  TVideoRendererOverlay = class(TVideoRendererClass)
  private
    pDDraw: IDirectDraw;
    pPrimary: IDirectDrawSurface;
    pOverlay: IDirectDrawSurface;
    pClipper: IDirectDrawClipper;

    // --- window we belong to
    pHWND: HWND;
    pVideoMode: TVideoMode;
    pPhysBpp: Byte;


    // --- blbosti na desktop play
    WallP: PWideChar;
    Pattern: PWideChar;
    BackC: DWord;
    DesktopMode: Boolean;

  public

    function Stop: HResult;override;
    function GetVideoMode: TVideoMode;override;

    function Draw(buffer: Pointer; rect: TRect; FrameNumber: Integer; InvertFlag: DWord): HResult;override;
    function UpdateFrame(rect: TRect): HResult;override;

    function InitDDraw(fHwnd: HWND; Width, Height: Integer): HResult;override;
    function IsSupported: Boolean;override;

    procedure ClearToBlack;override;
    procedure Close;override;
    procedure OnDesktopChange(Sender: TObject; fHwnd:HWND; cX, cY: Integer; bpp: Byte);override;

    function SetDesktopPlay(Desk: Boolean): HResult;override;
    procedure PrepareDesktopMode;
    procedure UnPrepareDesktopMode;


    destructor Destroy; override;
    constructor Create;override;
  end;

implementation

const CLSID_ActiveDesktop : TGUID = '{75048700-EF1F-11D0-9888-006097DEACF9}';

const
  FourCC_YUY2 = 844715353;
  FourCC_UYVY = 1498831189;
  FourCC_YV12 = 842094169;

  ddpfOverlayFormats: Array[0..2] of TDDPixelFormat =
           ( ( dwSize: sizeof(TDDPixelFormat);
               dwFlags: DDPF_FOURCC;
               dwFourCC: FourCC_YUY2;
               dwYUVBitCount: 0;
               dwYBitMask: 0;
               dwUBitMask: 0;
               dwVBitMask: 0;
               dwYUVAlphaBitMask: 0;
             ),
             ( dwSize: sizeof(TDDPixelFormat);
               dwFlags: DDPF_FOURCC;
               dwFourCC: FourCC_UYVY;
               dwYUVBitCount: 0;
               dwYBitMask: 0;
               dwUBitMask: 0;
               dwVBitMask: 0;
               dwYUVAlphaBitMask: 0;
             ),
             ( dwSize: sizeof(TDDPixelFormat);
               dwFlags: DDPF_FOURCC;
               dwFourCC: FourCC_YV12;
               dwYUVBitCount: 0;
               dwYBitMask: 0;
               dwUBitMask: 0;
               dwVBitMask: 0;
               dwYUVAlphaBitMask: 0;
             )
           );

procedure TVideoRendererOverlay.UnPrepareDesktopMode;
 var hr: HResult;
     pDesktop: IActiveDesktop;
     background: Integer;
begin
  hr:= CoCreateInstance(CLSID_ActiveDesktop, nil, CLSCTX_INPROC_SERVER, IID_IActiveDesktop, pDesktop);

  if pDesktop <> nil then begin
    pDesktop.SetWallpaper(Wallp, 0);
    pDesktop.SetPattern(Pattern, 0);
    pDesktop.ApplyChanges(AD_APPLY_ALL);
    pDesktop:= nil;
  end;

  background:= Color_Desktop;
  SetSysColors(1, background, BackC);
  SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, nil, 0);

  DesktopMode:= False;
end;

procedure TVideoRendererOverlay.PrepareDesktopMode;
 var hr: HResult;
     pDesktop: IActiveDesktop;
     Col: DWord;
     background: Integer;
begin
  hr:= CoCreateInstance(CLSID_ActiveDesktop, nil, CLSCTX_INPROC_SERVER, IID_IActiveDesktop, pDesktop);

  if pDesktop <> nil then begin
    pDesktop.GetWallpaper(Wallp, MAX_PATH, 0);
    pDesktop.GetPattern(Pattern, MAX_PATH, 0);

    pDesktop.SetWallpaper(PWideChar(WideString('None')), 0);
    pDesktop.SetPattern(PWideChar(WideString('None')), 0);
    pDesktop.ApplyChanges(AD_APPLY_ALL);
    pDesktop:= nil;
  end;

  Background:= Color_Desktop;
  BackC:= GetSysColor(COLOR_DESKTOP);
  SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(''), 0);
  Col:= Overlay_Color;
  SetSysColors(1, background, Col);

  DesktopMode:= True;
end;

function TVideoRendererOverlay.SetDesktopPlay(Desk: Boolean): HResult;
begin
  result:= S_OK;
  DesktopMode:= desk;

  if Desk then begin
    PrepareDesktopMode;
  end else begin
    UnPrepareDesktopMode;
  end;
end;

procedure TVideoRendererOverlay.OnDesktopChange(Sender: TObject; fHwnd: HWND; cX, cY: Integer; bpp: Byte);
begin
  if pDDraw <> nil then begin
     InitDDraw(fHwnd, cx, cy);
  end;
end;

procedure TVideoRendererOverlay.ClearToBlack;
 var x: TDDBLTFx;
begin
  if (pDDraw <> nil) and (pPrimary <> nil) then begin

    ZeroMemory(@x, sizeof(x));
    x.dwSize:= sizeof(x);
  case pPhysBpp of
    2 : x.dwFillColor:= OVERLAY_COLORKEY_16;
    3, 4: x.dwFillColor:= OVERLAY_COLORKEY_32;
  end;

    pPrimary.Blt(nil, nil, nil, DDBLT_COLORFILL or DDBLT_WAIT, @x);
  end;
end;

function TVideoRendererOverlay.Draw(buffer: Pointer; rect: TRect; FrameNumber: Integer; InvertFlag: DWord): HResult;
 var hr: HResult;
     srcRect, destRect: TRect;
     pt: TPoint;
     desc: TDDSurfaceDesc;
     UpdateFlags: DWord;

     i: Dword;
     pBDest, pBSrc: PByteArray;
     pCopyCount: DWord;
     pOffsSrc: DWord;
     pOffsDest: DWord;

     oFX: TDDOverlayFX;
     fx: TDDBltFx;

     scrW, scrH: Integer;
begin
  if (buffer = nil) or (pOverlay = nil) or (pPrimary = nil) then
            begin result:= V_ERROR; exit; end;

  ZeroMemory(@Desc, sizeof(desc));
  desc.dwSize:= sizeof(desc);

  hr:= pOverlay.Lock(nil, desc, DDLOCK_SurfacememoryPtr or DDLock_Wait, 0);
  if failed(hr) then begin result:= V_ERROR; exit; end;

  //  prepare all needed numbers so we don't need to do it in the CYCLE
  pBDest:= desc.lpSurface;
  pBSrc:= buffer;
  pCopyCount:= pWidth * pBPP;
  pOffsSrc:= 0;

  case InvertFlag of
    INVERT_NONE : begin
                    pOffsDest:= (desc.dwHeight-1) * desc.lPitch;
                    for i:= 0 to desc.dwHeight-1 do begin
                      Move( pBSrc^[pOffsSrc], pBDest^[pOffsDest], pCopyCount);
                      pOffsSrc:= pOffsSrc + pCopyCount;
                      pOffsDest:= pOffsDest - desc.lPitch;
                    end;
                  end;
    INVERT_UPDOWN: begin
                    pOffsDest:= 0;
                    for i:= 0 to desc.dwHeight-1 do begin
                      Move( pBSrc^[pOffsSrc], pBDest^[pOffsDest], pCopyCount);
                      pOffsSrc:= pOffsSrc + pCopyCount;
                      pOffsDest:= pOffsDest + desc.lPitch;
                    end;
                   end;
  end;

  pBSrc:= nil; pBDest:= nil;
  pOverlay.Unlock(nil);


  UpdateFlags:= DDOVER_SHOW or DDOVER_DDFX or DDOVER_KEYDESTOVERRIDE;

  ZeroMemory(@oFX, sizeof(oFX));
  oFX.dwSize:= sizeof(oFX);

  case pPhysBpp of
    2 : begin
          oFX.dckSrcColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_16;
          oFX.dckSrcColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_16;

          oFX.dckDestColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_16;
          oFX.dckDestColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_16;
        end;
     3, 4: begin
          oFX.dckSrcColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_32;
          oFX.dckSrcColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_32;

          oFX.dckDestColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_32;
          oFX.dckDestColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_32;
           end;
  end;

  scrW:= GetSystemMetrics(SM_CXSCREEN);
  scrH:= GetSystemMetrics(SM_CYScreen);

  srcRect.Left:= 0;
  srcRect.Top:= 0;
  srcRect.Right:= pWidth;
  srcRect.Bottom:= pHeight;

  if not DesktopMode then begin
    pt.X:= 0;
    pt.Y:= 0;
    ClientToScreen(pHWND, pt);

    destRect:= Rect;
    OffsetRect(DestRect, pt.x, pt.y);
  end else begin
    DestRect.Left:= 0;
    DestRect.Top:= 0;
    DestRect.Right:= scrW;
    DestRect.Bottom:= scrH;
  end;


  if (destRect.Right <= scrW) and (destRect.Left >=0) and
     (destRect.Top >=0) and (destRect.Bottom <= scrH) then
     begin

     end else begin

        if destRect.Right > scrW then begin
           srcRect.Right:=  Round(  (scrW - destRect.left) *  pWidth / (destRect.Right - destRect.Left) );
           destRect.Right:= scrW;
        end;

        if destRect.Left < 0 then begin
           srcRect.Left:= Round( -destRect.Left * pWidth / (destRect.Right - destRect.Left) );
           destRect.Left:= 0;
        end;

        if destRect.Bottom > scrH then begin
           srcRect.Bottom:= Round(  (scrH - destRect.Top) * pHeight / (destRect.Bottom-destRect.Top) );
           destRect.Bottom:= scrH;
        end;

        if destRect.Top < 0 then begin
           srcRect.Top:= Round( -destRect.Top * pHeight / (destRect.Bottom - destRect.Top) );
           destRect.Top:= 0;
        end;
     end;

  hr:= pOverlay.UpdateOverlay(@srcRect, pPrimary, @destRect, UpdateFlags, @oFX);
  if failed(hr) then begin result:= V_ERROR; exit; end;

  result:= NoError;
end;

function TVideoRendererOverlay.UpdateFrame(rect: TRect): HResult;
 var srcRect, destRect: TRect;
     pt: TPoint;
     UpdateFlags: DWord;
     scrW, scrH: Integer;
     hr: HResult;
     oFX: TDDOverlayFX;
begin
  UpdateFlags:= DDOVER_SHOW or DDOVER_DDFX or DDOVER_KEYDESTOVERRIDE;

  ZeroMemory(@oFX, sizeof(oFX));
  oFX.dwSize:= sizeof(oFX);

  case pPhysBpp of
    2 : begin
          oFX.dckSrcColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_16;
          oFX.dckSrcColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_16;

          oFX.dckDestColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_16;
          oFX.dckDestColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_16;
        end;
     3, 4: begin
          oFX.dckSrcColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_32;
          oFX.dckSrcColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_32;

          oFX.dckDestColorkey.dwColorSpaceLowValue:= OVERLAY_COLORKEY_32;
          oFX.dckDestColorkey.dwColorSpaceHighValue:= OVERLAY_COLORKEY_32;
           end;
  end;

  pt.X:= 0;
  pt.Y:= 0;
  ClientToScreen(pHWND, pt);

  destRect:= Rect;
  OffsetRect(DestRect, pt.x, pt.y);

  srcRect.Left:= 0;
  srcRect.Top:= 0;
  srcRect.Right:= pWidth;
  srcRect.Bottom:= pHeight;

  scrW:= GetSystemMetrics(SM_CXSCREEN);
  scrH:= GetSystemMetrics(SM_CYScreen);


  // --- obrezeme source rect ak precnievame...
  if (destRect.Right <= scrW) and (destRect.Left >=0) and
     (destRect.Top >=0) and (destRect.Bottom <= scrH) then
     begin

     end else begin

        if destRect.Right > scrW then begin
           srcRect.Right:=  Round(  (scrW - destRect.left) *  pWidth / (destRect.Right - destRect.Left) );
           destRect.Right:= scrW;
        end;

        if destRect.Left < 0 then begin
           srcRect.Left:= Round( -destRect.Left * pWidth / (destRect.Right - destRect.Left) );
           destRect.Left:= 0;
        end;

        if destRect.Bottom > scrH then begin
           srcRect.Bottom:= Round(  (scrH - destRect.Top) * pHeight / (destRect.Bottom-destRect.Top) );
           destRect.Bottom:= scrH;
        end;

        if destRect.Top < 0 then begin
           srcRect.Top:= Round( -destRect.Top * pHeight / (destRect.Bottom - destRect.Top) );
           destRect.Top:= 0;
        end;
     end;

  hr:= pOverlay.UpdateOverlay(@srcRect, pPrimary, @destRect, UpdateFlags, @oFX);
  if failed(hr) then begin result:= V_ERROR; exit; end;

  result:= NoError;
end;

function TVideoRendererOverlay.Stop: HResult;
 var ofx: TDDOverlayFX;
begin
  result:= NoError;
{  if pOverlay <> nil then begin
    ZeroMemory(@ofx, sizeof(ofx));
    ofx.dwSize:= sizeof(ofx);
    pOverlay.UpdateOverlay(nil, pPrimary, nil, DDOVER_HIDE, @ofx);
  end;}
end;

function TVideoRendererOverlay.GetVideoMode: TVideoMode;
begin
  result:= pVideoMode;
end;

function TVideoRendererOverlay.IsSupported: Boolean;
 var capsDrv: TDDCaps;
     hlpCaps: TDDCaps;
     hr: HResult;
begin
  ZeroMemory(@capsDrv, sizeof(capsDrv));
  capsDrv.dwSize:= sizeof(capsDrv);

  Zeromemory(@hlpCaps, sizeof(hlpCaps));
  hlpCaps.dwSize:= sizeof(hlpCaps);

  result:= True;

{  hr:= pDDraw.GetCaps(@capsDrv, @hlpCaps);
  if failed(hr) then begin result:= False; exit; end;

  if (capsDrv.dwCaps and DDCaps_Overlay) = 0 then result:= False
   else result:= True;} 
end;

function TVideoRendererOverlay.InitDDraw(fHwnd: HWND; Width, Height: Integer): HResult;
 var hr: HResult;
     desc, descO: TDDSurfaceDesc;
     i: Integer;
begin
  result:= V_ERROR;
  if ( Width <=0 ) or (Height <=0) then exit;

  pClipper:= nil;
  pPrimary:= nil;
  pOverlay:= nil;
  pDDraw:= nil;

  hr:= DirectDrawCreate(nil, pDDraw, nil);
  if failed(hr) then exit;

  /// --- keep the window handle
  pHWND:= fHwnd;

  if not IsSupported then begin result:= V_NO_OVERLAY; Close; exit; end;

  hr:= pDDraw.SetCooperativeLevel(pHwnd, DDSCL_NORMAL);
  if failed(hr) then begin Close; exit; end;

  ZeroMemory(@desc, sizeof(desc));
  desc.dwSize:= Sizeof(desc);
  desc.dwFlags:= DDSD_CAPS;
  desc.ddsCaps.dwCaps:= DDSCAPS_PRIMARYSURFACE;
  hr:= pDDraw.CreateSurface(desc, pPrimary, nil);
  if failed(hr) then begin Close; exit; end;

  hr:= pDDraw.CreateClipper(0, pClipper, nil);
  if failed(hr) then begin close; Exit; end;

  hr:= pClipper.SetHwnd(0, pHwnd);
  if failed(hr) then begin close; exit; end;

  hr:= pPrimary.SetClipper(pClipper);

  // --- creating Overlay Surface
  ZeroMemory(@descO, sizeof(descO));
  descO.dwSize:= sizeof(descO);
  descO.ddsCaps.dwCaps:= DDSCAPS_OVERLAY or DDSCAPS_VIDEOMEMORY;
  descO.dwFlags:= DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT;
  descO.dwWidth:= Width;
  descO.dwHeight:= Height;

  // --- let's try the overlay formats
  i:= 0;
  pVideoMode:= VIDEOMODE_NONE;
  repeat
    descO.ddpfPixelFormat:= ddpfOverlayFormats[i];
    hr:= pDDraw.CreateSurface(descO, pOverlay, nil);
    if succeeded(hr) then begin
      case i of
        0 : pVideoMode:= VIDEOMODE_YUY2;
        1 : pVideoMode:= VIDEOMODE_UYVY;
        2 : pVideoMode:= VIDEOMODE_YV12;
      end;
    end;

    inc(i);
  until  (Succeeded(hr)) or (i>2);

  if (i > 2) and (failed(hr)) then begin
    Close;
    result:= V_NO_OVERLAY;
    exit;
  end;

  pWidth:= Width;
  pHeight:= Height;
  pBPP:= 2;      // 16 bit -> 2 Bytes

  ZeroMemory(@desc, sizeof(desc));
  desc.dwSize:= sizeof(desc);

  hr:= pPrimary.Lock(nil, desc, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WAIT or DDLOCK_WRITEONLY, 0);
  if failed(hr) then begin result:= V_ERROR; exit; end;
  pPrimary.Unlock(nil);

  case desc.ddpfPixelFormat.dwRGBBitCount of
    8: begin result:= V_ERROR; exit; end;
    16: pPhysBpp:= 2;
    24: pPhysBpp:= 3;
    32: pPhysBpp:= 4;
  end;

  result:= NoError;
end;


procedure TVideoRendererOverlay.Close;
begin
  if pDDraw <> nil then pDDraw.SetCooperativeLevel(pHWND, DDSCL_NORMAL);
  pClipper:= nil;
  pOverlay:= nil;
  pPrimary:= nil;
  pDDraw:= nil;

  pVideoMode:= VIDEOMODE_NONE;
  pBPP:= 0;
end;

destructor TVideoRendererOverlay.Destroy;
begin
  Close;
  if desktopmode then SetDesktopPlay(false);
  inherited;
end;

constructor TVideoRendererOverlay.Create;
begin
  inherited;

  pVideoMode:= VIDEOMODE_NONE;
  pBPP:= 0;
  pPhysBpp:= 0;

  pDDraW:= nil;
  pPrimary:= nil;
  pOverlay:= nil;
  pClipper:= nil;
end;

end.
